#include <assert.h>
#include "hdnum.hh"   
/** @brief Explicit Runge Kutta method as an example for an ODE solver

    The ODE solver is parametrized by a model. The model also
    exports all relevant types for time and states.
    The ODE solver encapsulates the states needed for the computation.

    Every explicit RK method can be realized by providing an
    appropriate Butcher Tableau. However, some tableaus are already
    hard coded and therefore integrators of order one to four can be
    chosen by using the second constructor and simply passing the
    desired order.

    \tparam M the model type
*/

using namespace hdnum; // Namen ohne hdnum:: verwenden 

template<class M>
class ExplicitRungeKutta
{
public:
  /** \brief export size_type */
  typedef typename M::size_type size_type;

  /** \brief export time_type */
  typedef typename M::time_type time_type;

  /** \brief export number_type */
  typedef typename M::number_type number_type;

  /** \brief the type of a Butcher tableau */
  typedef DenseMatrix<number_type> ButcherTableau;

private:
  
  //! Some default tableaus for the standard RK methods of order one
  //! to four.
  static ButcherTableau initTableau(const size_type order)
  {
    if(order == 4){
      ButcherTableau butcher(5,5,0.0);
      butcher[1][0] = 0.5;  butcher[1][1] = 0.5;
      butcher[2][0] = 0.5;  butcher[2][2] = 0.5;
      butcher[3][0] = 1;    butcher[3][3] = 1;
      butcher[4][1] = 1/6.; butcher[4][2] = 2/6.; 
      butcher[4][3] = 2/6.; butcher[4][4] = 1/6.;
      return butcher;
    }
    else if(order == 3){
      ButcherTableau butcher(4,4,0.0);
      butcher[1][0] = 0.5;  butcher[1][1] = 0.5;
      butcher[2][0] = 1;    butcher[2][1] = -1;    butcher[2][2] = 2;
      butcher[3][1] = 1./6.;butcher[3][2] = 4./6.; butcher[3][3] = 1./6.;
      return butcher;
    }
    else if(order == 2){
      ButcherTableau butcher(3,3,0.0);
      butcher[1][0] = 0.5; butcher[1][1] = 0.5;
      butcher[2][2] = 1;
      return butcher;
    }
    else if(order == 1){
      ButcherTableau butcher(2,2,0.0);
      butcher[1][1] = 1;
      return butcher;
    }
    else{
      HDNUM_ERROR("Order not available for Runge Kutta solver.");
    }
  }

public:

  //! Constructor requires a reference to the underlying model and the
  //! order of Runge-Kutta integration. Orders one to four are
  //! available. But arbitrary explicit RK methods can be realized
  //! with the alternative constructor.
  ExplicitRungeKutta (const M& model_, const size_type _order)
    : model(model_), butcher(initTableau(_order)), u(model.size()), f(model.size()), order(_order)
  {
    model.initialize(t,u);
    dt = 0.1; 
  }

  /** Constructor requires a reference to the underlying model. The
      method is defined by passing a butcher tableau of arbitrary size
      according to the pattern:

      c_1 | 0
      c_2 | a_{2,1}  0
       .  |  .             .
       .  |  .                  .
       .  |  .                       .
      c_s | a_{s,1}  .  .  . a_{s,s-1}  0
      -----------------------------------
          | b_1      .  .  .            b_s

     The RK method is then defined by
     
     y_n = y_{n-1} + h_n (b_1 k_1 + ... + b_s k_s)

     where
     
     k_i = f(t_{n-1} + c_i h_n, y_{n-1} 
         + h_n \sum_{j=1}^{i-1} a_{i,j} k_j).

     The constructor also requires knowledge about the order of the
     method, which is important if this solver is applied within some
     e.g.  Runge-Kutta-Fehlberg Algorithm.
   */
  ExplicitRungeKutta (const M& model_, const DenseMatrix<number_type> & butcher_, const size_type _order)
    : model(model_), butcher(butcher_), u(model.size()), f(model.size()), order(_order)
  {
    assert(butcher.rowsize() == butcher.colsize());
    model.initialize(t,u);
    dt = 0.1; 
  }

  //! set time step for subsequent steps
  void set_dt (time_type dt_)
  {
    dt = dt_;
  }

  void set_verbosity (int v)
  {
    
  }

  void get_info () const
  {
  }

  size_type get_order() const
  {
    return order;
  }

  //! do one step
  void step ()
  {
    const size_type R = butcher.colsize()-1;

    std::vector< Vector<number_type> > k;
    for(size_type i=0; i<R; ++i)
      k.push_back( Vector<number_type>( model.size()) );

    for(size_type i=0; i<R; ++i){
      const number_type t_i = t + butcher[i][0] * dt;

      Vector<number_type> u_i(u);
      for(size_type j=0; j<i; ++j)
        u_i.update(dt * butcher[i][j+1],k[j]);

      k[i] = 0;
      model.f(t_i,u_i,k[i]);
    }

    for(size_type i=0; i<R; ++i)
      u.update(dt * butcher[R][i+1], k[i]);

    t += dt;
  }

  //! get current state
  const Vector<number_type>& get_state () const
  {
    return u;
  }

  //! get current time
  time_type get_time () const
  {
    return t;
  }

  //! get dt used in last step (i.e. to compute current state)
  time_type get_dt () const
  {
    return dt;
  }

private:
  const M& model;                    
  const DenseMatrix<number_type> butcher; 
  time_type t, dt;
  Vector<number_type> u;
  Vector<number_type> f;
  size_type order;
};
